# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import io
import json
import unicodecsv as csv
from tabulator import topen
from importlib import import_module

from .model import SchemaModel
from . import compat
from . import utilities


# Module API

def push_resource(table, schema, data, backend, **backend_options):
    """Push JSONTableSchema resource to storage's table.

    All parameters should be used as keyword arguments.

    Parameters
    ----------
    table: str
        Table name.
    schema: str
        Path to schema file.
    data: str
        Path to data file.
    backend: str
        Backend name like `sql` or `bigquery`.
    backend_options: dict
        Backend options mentioned in backend docs.

    """

    # Get storage
    plugin = import_module('jsontableschema.plugins.%s' % backend)
    storage = plugin.Storage(**backend_options)

    # Create table
    model = SchemaModel(schema)
    schema = model.as_python
    if storage.check(table):
        storage.delete(table)
    storage.create(table, schema)

    # Write data
    with topen(data, with_headers=True) as data:
        storage.write(table, data)


def pull_resource(table, schema, data, backend, **backend_options):
    """Pull JSONTableSchema resource from storage's table.

    All parameters should be used as keyword arguments.

    Parameters
    ----------
    table: str
        Table name.
    schema: str
        Path to schema file.
    data: str
        Path to data file.
    backend: str
        Backend name like `sql` or `bigquery`.
    backend_options: dict
        Backend options mentioned in backend docs.

    """

    # Get storage
    plugin = import_module('jsontableschema.plugins.%s' % backend)
    storage = plugin.Storage(**backend_options)

    # Save schema
    mode = 'w'
    encoding = 'utf-8'
    if compat.is_py2:
        mode = 'wb'
        encoding = None
    utilities.ensure_dir(schema)
    with io.open(schema, mode=mode, encoding=encoding) as file:
        schema = storage.describe(table)
        json.dump(schema, file, indent=4)

    # Save data
    utilities.ensure_dir(data)
    with io.open(data, 'wb') as file:
        model = SchemaModel(schema)
        data = storage.read(table)
        writer = csv.writer(file, encoding='utf-8')
        writer.writerow(model.headers)
        for row in data:
            writer.writerow(row)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from functools import partial
from abc import ABCMeta, abstractmethod
from .. import compat
from .. import exceptions
from .. import constraints


# Module API

class JTSType(object):
    """Base class for all JSON Table Schema types.

    Args:
        field (dict): field schema

    Aside implementing `cast_default` subclass could add other cast methods
    with the same signature like `cast_fmt`, `cast_current` etc
    to add support for corresponding formats.

    """

    # Public

    @property
    @abstractmethod
    def name(self):
        """str: type name like `array`
        """
        pass  # pragma: no cover

    @property
    @abstractmethod
    def null_values(self):
        """str[]: list of strings to process as null value
        """
        pass  # pragma: no cover

    @property
    @abstractmethod
    def supported_constraints(self):
        """str[]: list of supported JTS constraints
        """
        pass  # pragma: no cover

    def __init__(self, field=None):

        # Set default field
        self.__field = None
        self.__field_name = None
        self.__format = 'default'
        self.__constraints = {}

        # Set user defined field
        if field:
            self.__field = field
            self.__field_name = field['name']
            self.__format = field['format']
            self.__constraints = field.get('constraints', {})

        # Set parsed format (fmt feature)
        self.__format_main = self.__format
        self.__format_fmt = None
        if self.__format.startswith('fmt'):
            self.__format_main = 'fmt'
            self.__format_fmt = self.__format.strip('fmt:')

    def cast(self, value, skip_constraints=False):
        """Cast value.

        Args:
            value (any): value to cast
            skip_constraints (bool): if True it skips constraints checks

        Returns:
           any: cast value

        """

        # If value is null
        if value in self.null_values + [None]:

            # Check required constraint
            if not skip_constraints:
                required = self.__constraints.get('required', False)
                constraints.check_required(self.__field_name, value, required,
                    self.null_values+[None])

            return None

        # Check pattern constraint
        if not skip_constraints:
            # Only if value not cast
            if isinstance(value, compat.str):
                pattern = self.__constraints.get('pattern', None)
                if pattern is not None:
                    constraints.check_pattern(
                        self.__field_name, value, pattern)

        # Cast value
        cast_name = 'cast_%s' % self.__format_main
        cast_func = getattr(self, cast_name, self.cast_default)
        cast_value = cast_func(value, self.__format_fmt)

        # Check against post-cast constraints
        if not skip_constraints:
            for check_name, check_value in self.__constraints.items():
                if check_name in ['required', 'pattern']:
                    continue
                if check_name not in self.supported_constraints:
                    raise exceptions.ConstraintNotSupported(
                        "Field type '{0}' does not support the {1} constraint"
                        .format(self.name, check_name))
                if check_name in ['minimum', 'maximum']:
                    check_value = self.cast(check_value, skip_constraints=True)
                if check_name in ['enum']:
                    mapper = partial(self.cast, skip_constraints=True)
                    check_value = map(mapper, check_value)
                check_func = getattr(constraints, 'check_%s' % check_name)
                check_func(self.__field_name, cast_value, check_value)

        return cast_value

    def test(self, value):
        """Test value could be cast.

        Args:
            value (any): value to check

        Returns:
            bool: could be cast

        """
        try:
            self.cast(value)
            return True
        except exceptions.InvalidCastError:
            return False

    @abstractmethod
    def cast_default(self, value, fmt=None):
        """Cast default.

        Args:
            value (any): value to cast
            format (str): secondary format (JTS's "fmt")

        """
        pass  # pragma: no cover



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import datetime
from future.utils import raise_with_traceback
from dateutil.parser import parse as date_parse
from .. import exceptions
from .. import utilities
from . import base


# Module API

class DateTimeType(base.JTSType):

    # Public

    name = 'datetime'
    null_values = utilities.NULL_VALUES
    supported_constraints = [
        'required',
        'pattern',
        'enum',
        'minimum',
        'maximum',
    ]
    # ---
    python_type = datetime.datetime
    ISO8601 = '%Y-%m-%dT%H:%M:%SZ'
    raw_formats = ['DD/MM/YYYYThh/mm/ss']
    py_formats = ['%Y/%m/%dT%H:%M:%S']
    format_map = dict(zip(raw_formats, py_formats))

    def cast_default(self, value, fmt=None):

        if isinstance(value, self.python_type):
            return value

        try:
            return datetime.datetime.strptime(value, self.ISO8601)
        except (TypeError, ValueError) as e:
            raise_with_traceback(exceptions.InvalidDateTimeType(e))

    def cast_any(self, value, fmt=None):

        if isinstance(value, self.python_type):
            return value

        try:
            return date_parse(value)
        except (TypeError, ValueError) as e:
            raise_with_traceback(exceptions.InvalidDateTimeType(e))

    def cast_fmt(self, value, fmt=None):

        if isinstance(value, self.python_type):
            return value

        try:
            return datetime.datetime.strptime(value, fmt)
        except (TypeError, ValueError) as e:
            raise_with_traceback(exceptions.InvalidDateTimeType(e))



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .any import AnyType
from .array import ArrayType
from .base import JTSType
from .boolean import BooleanType
from .date import DateType
from .datetime import DateTimeType
from .geojson import GeoJSONType
from .geopoint import GeoPointType
from .integer import IntegerType
from .null import NullType
from .number import NumberType
from .object import ObjectType
from .string import StringType
from .time import TimeType



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from . import exceptions
from . import types
from . import model
from . import storage
from .infer import infer
from .pushpull import push_resource, pull_resource
from .validate import validate, validator


__all__ = ['exceptions', 'types', 'model', 'utilities',
           'compat', 'infer', 'validate', 'validator',
           'storage', 'push_resource', 'pull_resource']



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import re
from .. import exceptions
from .. import compat


# Module API

def check_pattern(name, value, pattern):
    """Pattern constraint for a string value.

    Supported types: all.
    Input arguments should NOT be casted to type.
    Pattern constraint should be checked as a string value before the value
    is cast. `value` is treated as a string and must match the XML Schema Reg
    Exp `pattern`.

    Args:
        name (str): field name
        value (str): field value
        pattern (str): pattern to check

    Raises:
        TypeError: for non supported type
        ConstraintError: if check is failed

    """

    # Convert to str if necessary
    if not isinstance(value, compat.str):
        value = compat.str(value)

    # Check constaint
    regex = re.compile('^{0}$'.format(pattern))
    match = regex.match(value)
    if not match:
        raise exceptions.ConstraintError(
            "The value for field '{0}' must match the pattern"
            .format(name))

    return True



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .. import exceptions


# Module API

def check_unique(name, values):
    """Unique values constraint.

    Raises:
        ConstraintNotSupported: always

    """
    raise exceptions.ConstraintNotSupported(
        'Unique constraint is not supported')



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import json

import jsonschema
from jsonschema.validators import validator_for

from . import compat, exceptions


def load_validator():
    filepath = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                            'schemas/json-table-schema.json')
    with open(filepath) as f:
        json_table_schema = json.load(f)
    return json_table_schema, validator_for(json_table_schema)


json_table_schema, BaseValidator = load_validator()


class JSONTableSchemaValidator(BaseValidator):
    @classmethod
    def check_schema(cls, schema):
        # When checking against the metaschema, we do not want to run the
        # additional checking added in iter_errors
        parent_cls = cls.__bases__[0]
        for error in parent_cls(cls.META_SCHEMA).iter_errors(schema):
            raise jsonschema.exceptions.SchemaError.create_from(error)

    def iter_errors(self, instance, _schema=None):
        for e in super(JSONTableSchemaValidator, self).iter_errors(instance,
                                                                   _schema):
            yield exceptions.SchemaValidationError(
                e.message, e.validator, e.path, e.cause, e.context,
                e.validator_value, e.instance, e.schema, e.schema_path,
                e.parent
            )

        try:
            field_names = [f['name'] for f in instance['fields']]
        except (TypeError, KeyError):
            field_names = []
        # the hash MAY contain a key `primaryKey`
        if isinstance(instance, dict) and instance.get('primaryKey'):
            # ensure that the primary key matches field names

            if isinstance(instance['primaryKey'], compat.str):
                if not instance['primaryKey'] in field_names:
                    yield exceptions.SchemaValidationError(
                        'A JSON Table Schema primaryKey value must be found in'
                        ' the schema field names'
                    )
            elif isinstance(instance['primaryKey'], list):
                for k in instance['primaryKey']:
                    if k not in field_names:
                        yield exceptions.SchemaValidationError(
                            'A JSON Table Schema primaryKey value must be '
                            'found in the schema field names'
                        )

        # the hash may contain a key `foreignKeys`
        if isinstance(instance, dict) and instance.get('foreignKeys'):
            for fk in instance['foreignKeys']:
                # ensure that `foreignKey.fields` match field names
                if isinstance(fk.get('fields'), compat.str):
                    if fk.get('fields') not in field_names:
                        yield exceptions.SchemaValidationError(
                            'A JSON Table Schema foreignKey.fields value must '
                            'correspond with field names.'
                        )
                elif isinstance(fk.get('fields', []), list):
                    for field in fk.get('fields'):
                        if field not in field_names:
                            yield exceptions.SchemaValidationError(
                                'A JSON Table Schema foreignKey.fields value '
                                'must correspond with field names.'
                            )

                # ensure that `foreignKey.reference.fields`
                # matches outer `fields`
                if isinstance(fk.get('fields'), compat.str):
                    if not isinstance(fk['reference']['fields'], compat.str):
                        yield exceptions.SchemaValidationError(
                            'A JSON Table Schema foreignKey.reference.fields '
                            'must match field names.'
                        )
                else:
                    if isinstance(fk['reference']['fields'], compat.str):
                        yield exceptions.SchemaValidationError(
                            'A JSON Table Schema foreignKey.fields cannot '
                            'be a string when foreignKey.reference.fields.'
                            'is a string'
                        )
                    if not len(fk.get('fields')) == len(fk['reference']['fields']):
                        yield exceptions.SchemaValidationError(
                            'A JSON Table Schema foreignKey.fields must '
                            'contain the same number entries as '
                            'foreignKey.reference.fields.'
                        )


validator = JSONTableSchemaValidator(json_table_schema)


def validate(schema):
    jsonschema.validate(schema, json_table_schema,
                        cls=JSONTableSchemaValidator)
    return True



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import io
import sys
import json
import requests
from copy import deepcopy
from importlib import import_module

from . import compat
from . import exceptions


REMOTE_SCHEMES = ('http', 'https', 'ftp', 'ftps')
NULL_VALUES = ['null', 'none', 'nil', 'nan', '-', '']
TRUE_VALUES = ['yes', 'y', 'true', 't', '1', 1]
FALSE_VALUES = ['no', 'n', 'false', 'f', '0', 0]


def load_json_source(source):
    """Load a JSON source, from string, URL or buffer, into a Python type.
    """

    if source is None:
        return None

    elif isinstance(source, (dict, list)):
        # the source has already been loaded
        return deepcopy(source)

    if compat.parse.urlparse(source).scheme in REMOTE_SCHEMES:
        source = requests.get(source).text

    elif isinstance(source, compat.str) and not os.path.exists(source):
        pass

    else:
        with io.open(source, encoding='utf-8') as stream:
            source = stream.read()

    try:
        return json.loads(source)
    except ValueError as e:
        raise exceptions.InvalidJSONError


def ensure_dir(path):
    """Ensure directory exists.

    Parameters
    ----------
    path: str

    """
    dirpath = os.path.dirname(path)
    if dirpath and not os.path.exists(dirpath):
        os.makedirs(dirpath)


class PluginImporter(object):
    """Plugin importer.

    Example
    -------
    Add to myapp.plugins something like that::
        importer = PluginImporter(virtual='myapp.plugins.', actual='myapp_')
        importer.register()
        del PluginImporter
        del importer

    """

    # Public

    def __init__(self, virtual, actual):
        self.__virtual = virtual
        self.__actual = actual

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return (self.virtual == other.virtual and
                self.actual == other.actual)

    @property
    def virtual(self):
        return self.__virtual

    @property
    def actual(self):
        return self.__actual

    def register(self):
        if self not in sys.meta_path:
            sys.meta_path.append(self)

    def find_module(self, fullname, path=None):
        if fullname.startswith(self.virtual):
            return self
        return None

    def load_module(self, fullname):
        if fullname in sys.modules:
            return sys.modules[fullname]
        if not fullname.startswith(self.virtual):
            raise ImportError(fullname)
        realname = fullname.replace(self.virtual, self.actual)
        try:
            module = import_module(realname)
        except ImportError:
            message = 'Plugin "%s" is not installed. '
            message += 'Run `pip install %s` to install.'
            message = message % (fullname, realname)
            raise ImportError(message)
        sys.modules[realname] = module
        sys.modules[fullname] = module
        return module



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import json
import jsonschema
from future.utils import raise_with_traceback
from .. import exceptions
from .. import utilities
from .. import compat
from . import base


# Module API

class GeoJSONType(base.JTSType):

    # Public

    name = 'geojson'
    null_values = utilities.NULL_VALUES
    supported_constraints = [
        'required',
        'pattern',
        'enum',
    ]
    # ---
    python_type = dict
    spec = {
        'types': ['Point', 'MultiPoint', 'LineString', 'MultiLineString',
                  'Polygon', 'MultiPolygon', 'GeometryCollection', 'Feature',
                  'FeatureCollection']
    }

    def cast_default(self, value, fmt=None):
        if isinstance(value, self.python_type):
            try:
                jsonschema.validate(value, _geojson_schema)
                return value
            except jsonschema.exceptions.ValidationError as e:
                raise_with_traceback(exceptions.InvalidGeoJSONType())
        if isinstance(value, compat.str):
            try:
                geojson = json.loads(value)
                jsonschema.validate(geojson, _geojson_schema)
                return geojson
            except (TypeError, ValueError) as e:
                raise_with_traceback(exceptions.InvalidGeoJSONType())
            except jsonschema.exceptions.ValidationError as e:
                raise_with_traceback(exceptions.InvalidGeoJSONType())

    def cast_topojson(self, value, fmt=None):
        raise NotImplementedError


# Internal

def _load_geojson_schema():
    filepath = os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        '..', 'schemas', 'geojson.json')
    with open(filepath) as f:
        json_table_schema = json.load(f)
    return json_table_schema

_geojson_schema = _load_geojson_schema()



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .. import exceptions


# Module API

def check_required(name, value, required, null_values):
    """Required value constraint.

    Supported types: all.

    Args:
        name (str): field name
        value (str): field value
        null_values (str[]): list of null values

    Raises:
        TypeError: for non supported type
        ConstraintError: if check is failed

    """
    if required and value in null_values:
        message = 'The field "{0}" requires a value'
        message = message.format(name)
        raise exceptions.ConstraintError(message)
    return True



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import json
from .. import exceptions
from .. import utilities
from . import base


# Module API

class ArrayType(base.JTSType):

    # Public

    name = 'array'
    null_values = utilities.NULL_VALUES
    supported_constraints = [
        'required',
        'pattern',
        'enum',
        'minLength',
        'maxLength',
    ]
    # ---
    python_type = list

    def cast_default(self, value, fmt=None):

        if isinstance(value, self.python_type):
            return value

        try:
            array_type = json.loads(value)
            if isinstance(array_type, self.python_type):
                return array_type
            else:
                raise exceptions.InvalidArrayType('Not an array')

        except (TypeError, ValueError):
            raise exceptions.InvalidArrayType(
                '"{0}" is not a array type'.format(value))



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import re
import decimal
import unicodedata
from future.utils import raise_with_traceback
from .. import exceptions
from .. import compat
from .. import utilities
from . import base


# Module API

class NumberType(base.JTSType):

    # Public

    name = 'number'
    null_values = utilities.NULL_VALUES
    supported_constraints = [
        'required',
        'pattern',
        'enum',
        'minimum',
        'maximum',
    ]
    # ---
    python_type = decimal.Decimal
    separators = ',; '
    currencies = u''.join(compat.chr(i) for i
                          in range(0xffff)
                          if unicodedata.category(compat.chr(i)) == 'Sc')


    def cast_default(self, value, fmt=None):

        if isinstance(value, self.python_type):
            return value

        try:
            return self.python_type(value)
        except (ValueError, TypeError, decimal.InvalidOperation) as e:
            raise_with_traceback(exceptions.InvalidCastError(e))

        raise exceptions.InvalidCastError('Could not cast value')

    def cast_currency(self, value, fmt=None):

        if isinstance(value, self.python_type):
            return value

        if isinstance(value, (int, float)):
            return self.python_type(value)

        try:
            pattern = '[{0}{1}]'.format(self.separators, self.currencies)
            value = re.sub(pattern, '', value)
            return decimal.Decimal(value)
        except decimal.InvalidOperation:
            raise exceptions.InvalidCurrency(
                '{0} is not a valid currency'.format(value))

        raise exceptions.InvalidCastError('Could not cast value')



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .. import exceptions
from .. import utilities
from . import base


# Module API

class NullType(base.JTSType):

    # Public

    name = 'null'
    null_values = utilities.NULL_VALUES
    supported_constraints = [
        'required',
        'pattern',
        'enum',
    ]
    # ---
    python_type = type(None)

    def cast_default(self, value, fmt=None):

        if isinstance(value, self.python_type):
            return value

        value = value.strip().lower()
        if value in self.null_values:
            return None
        else:
            raise exceptions.InvalidNoneType(
                '{0} is not a none type'.format(value))



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from six import add_metaclass
from abc import ABCMeta, abstractmethod


# Module API

@add_metaclass(ABCMeta)
class Storage(object):
    """Tabular Storage Interface.

    Parameters
    ----------
    prefix: str
        Prefix for all tables.
    options: dict
        Concrete implementations options.

    """

    # Public

    @abstractmethod
    def __init__(self, prefix='', **options):
        pass  # pragma: no cover

    @property
    def tables(self):
        """Return list of storage's table names.

        Returns
        -------
        str[]
            List of table names.

        """
        pass  # pragma: no cover

    @abstractmethod
    def check(self, table):
        """Return true if table exists.

        Parameters
        ----------
        table: str
            Table name.

        Returns
        -------
        bool
            Table existence.

        """
        pass  # pragma: no cover

    @abstractmethod
    def create(self, table, schema):
        """Create table by schema.

        Parameters
        ----------
        table: str/list
            Table name or list of table names.
        schema: dict/list
            JSONTableSchema schema or list of schemas.

        Raises
        ------
        RuntimeError
            If table already exists.

        """
        pass  # pragma: no cover

    def delete(self, table):
        """Delete table.

        Parameters
        ----------
        table: str/list
            Table name or list of table names.

        Raises
        ------
        RuntimeError
            If table doesn't exist.

        """

    def describe(self, table):
        """Return table's JSONTableSchema schema.

        Parameters
        ----------
        table: str
            Table name.

        Returns
        -------
        dict
            JSONTableSchema schema.

        """

    def read(self, table):
        """Read data from table.

        Parameters
        ----------
        table: str
            Table name.

        Returns
        -------
        iterable
            Data tuples iterable.

        """

    def write(self, table, data):
        """Write data to table.

        Parameters
        ----------
        table: str
            Table name.
        data: iterable
            Iterable of data tuples.

        """
